"""App Tasks"""

# Standard Library
import datetime

# Third Party
# pylint: disable=no-name-in-module
from celery import shared_task

from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenExpiredError
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership, EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.tasks import QueueOnce

from ledger.decorators import when_esi_is_available
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.models.planetary import CharacterPlanetDetails
from ledger.task_helpers.char_helpers import (
    update_character_mining,
    update_character_wallet,
)
from ledger.task_helpers.core_helpers import enqueue_next_task, no_fail_chain
from ledger.task_helpers.corp_helpers import update_corp_wallet_division
from ledger.task_helpers.plan_helpers import (
    update_character_planetary,
    update_character_planetary_details,
)

logger = get_extension_logger(__name__)


@shared_task
@when_esi_is_available
def update_all_characters(runs: int = 0):
    characters = CharacterAudit.objects.select_related("character").all()
    for char in characters:
        update_character.apply_async(args=[char.character.character_id])
        runs = runs + 1
    logger.info("Queued %s Char Audit Updates", runs)


@shared_task(bind=True, base=QueueOnce)
def update_character(
    self, char_id, force_refresh=False
):  # pylint: disable=unused-argument
    character = (
        CharacterAudit.objects.select_related("character")
        .filter(character__character_id=char_id)
        .first()
    )
    if character is None:
        token = Token.get_token(char_id, CharacterAudit.get_esi_scopes())
        if token:
            try:
                if token.valid_access_token():
                    character, _ = CharacterAudit.objects.update_or_create(
                        character=EveCharacter.objects.get_character_by_id(
                            token.character_id
                        )
                    )
                else:
                    return False
            except TokenExpiredError:
                return False
        else:
            logger.info("No Tokens for %s", char_id)
            return False

    logger.debug(
        "Processing Audit Updates for %s", format(character.character.character_name)
    )

    skip_date = timezone.now() - datetime.timedelta(hours=1)
    que = []
    mindt = timezone.now() - datetime.timedelta(days=90)

    if (character.last_update_mining or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_mining_ledger.si(
                character.character.character_id, force_refresh=force_refresh
            )
        )

    if (character.last_update_wallet or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_wallet.si(
                character.character.character_id, force_refresh=force_refresh
            )
        )

    if (character.last_update_planetary or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_planets.si(
                character.character.character_id, force_refresh=force_refresh
            )
        )

    enqueue_next_task(que)

    logger.debug(
        "Queued %s Audit Updates for %s",
        len(que) + 1,
        character.character.character_name,
    )

    return True


# pylint: disable=unused-argument, too-many-locals
@shared_task(bind=True, base=QueueOnce)
def check_planetary_alarms(self, runs: int = 0):
    all_planets = CharacterPlanetDetails.objects.all()
    owner_ids = {}
    warnings = {}

    for planet in all_planets:
        if planet.is_expired and not planet.notification_sent and planet.notification:
            character_id = planet.planet.character.character.character_id

            # Determine if the character_id is part of any main character's alts
            main_id = None
            for main, alts in owner_ids.items():
                if character_id in alts:
                    main_id = main
                    break

            if main_id is None:
                try:
                    owner = CharacterOwnership.objects.get(
                        character__character_id=character_id
                    )
                    main = owner.user.profile.main_character
                    alts = main.character_ownership.user.character_ownerships.all()

                    owner_ids[main.character_id] = alts.values_list(
                        "character__character_id", flat=True
                    )

                    main_id = main.character_id
                except CharacterOwnership.DoesNotExist:
                    continue
                except AttributeError:
                    continue

            msg = _("%(charname)s on %(planetname)s") % {
                "charname": planet.planet.character.character.character_name,
                "planetname": planet.planet.planet.name,
            }

            if main_id not in warnings:
                warnings[main_id] = []

            warnings[main_id].append(msg)
            planet.notification_sent = True
            planet.save()

    if warnings:
        for main_id, messages in warnings.items():
            owner = CharacterOwnership.objects.get(character__character_id=main_id)
            title = _("Planetary Extractor Heads Expired")

            # Split messages into chunks of 50
            for i in range(0, len(messages), 50):
                chunk = messages[i : i + 50]
                msg = "\n".join(chunk)

                full_message = format_html(
                    "Following Planet Extractor Heads have expired: \n{}", msg
                )
                notify(
                    title=title,
                    message=full_message,
                    user=owner.user,
                    level="warning",
                )
                runs = runs + 1
    logger.info("Queued %s Planetary Alarms.", runs)


# Corporation Audit - Tasks


@shared_task
@when_esi_is_available
def update_all_corps(runs: int = 0):
    corps = CorporationAudit.objects.select_related("corporation").all()
    for corp in corps:
        update_corp.apply_async(args=[corp.corporation.corporation_id])
        runs = runs + 1
    logger.info("Queued %s Corp Audit Updates", runs)


@shared_task(bind=True, base=QueueOnce)
def update_corp(self, corp_id, force_refresh=False):  # pylint: disable=unused-argument
    corp = CorporationAudit.objects.get(corporation__corporation_id=corp_id)
    logger.debug("Processing Audit Updates for %s", corp.corporation.corporation_name)
    que = []

    que.append(update_corp_wallet.si(corp_id, force_refresh=force_refresh))

    enqueue_next_task(que)

    logger.debug("Queued Audit Updates for %s", corp.corporation.corporation_name)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["corp_id"]},
    name="tasks.update_corp_wallet",
)
@no_fail_chain
def update_corp_wallet(
    self, corp_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_corp_wallet_division(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_mining_ledger",
)
@no_fail_chain
def update_char_mining_ledger(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_mining(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_wallet",
)
@no_fail_chain
def update_char_wallet(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_wallet(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_planets",
)
@no_fail_chain
def update_char_planets(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_planetary(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id", "planet_id"]},
    name="tasks.update_char_planets_details",
)
@no_fail_chain
def update_char_planets_details(
    self, character_id, planet_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_planetary_details(
        character_id, planet_id, force_refresh=force_refresh
    )



noData = [
    {
        "ratting": [],
        "total": {
            "total_amount": "0.00",
            "total_amount_ess": "0.00",
            "total_amount_all": "0.00",
            "total_amount_mining": "0.00",
            "total_amount_others": "0.00",
            "total_amount_costs": "0.00",
        },
    }
]

CharmonthlyMarch = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "entity_type": "character",
                "total_amount": "200000.00",
                "total_amount_ess": "1133333.33",
                "total_amount_mining": "200000",
                "total_amount_others": "200000.00",
                "total_amount_costs": "-400000.00",
            }
        ],
        "total": {
            "total_amount": "200000.00",
            "total_amount_ess": "1133333.33",
            "total_amount_all": "1333333.33",
            "total_amount_mining": "200000.00",
            "total_amount_others": "200000.00",
            "total_amount_costs": "-400000.00",
        },
    }
]

CharmonthlyMarchWithTaxEvent = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "entity_type": "character",
                "total_amount": "200000.00",
                "total_amount_ess": "0.00",
                "total_amount_mining": "200000",
                "total_amount_others": "200000.00",
                "total_amount_costs": "-400000.00",
            }
        ],
        "total": {
            "total_amount": "200000.00",
            "total_amount_ess": "0.00",
            "total_amount_all": "200000.00",
            "total_amount_mining": "200000.00",
            "total_amount_others": "200000.00",
            "total_amount_costs": "-400000.00",
        },
    }
]

CharmonthlyMarchMulti = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "entity_type": "character",
                "total_amount": "200000.00",
                "total_amount_ess": "1133333.33",
                "total_amount_mining": "200000",
                "total_amount_others": "200000.00",
                "total_amount_costs": "-400000.00",
            },
            {
                "main_id": 1002,
                "main_name": "rotze Rotineque",
                "entity_type": "character",
                "total_amount": "0.00",
                "total_amount_ess": "0.00",
                "total_amount_mining": "100000",
                "total_amount_others": "0.00",
                "total_amount_costs": "0.00",
            },
        ],
        "total": {
            "total_amount": "200000.00",
            "total_amount_ess": "1133333.33",
            "total_amount_all": "1433333.33",
            "total_amount_mining": "300000.00",
            "total_amount_others": "200000.00",
            "total_amount_costs": "-400000.00",
        },
    }
]

Charyearly = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "entity_type": "character",
                "total_amount": "300000.00",
                "total_amount_ess": "1133333.33",
                "total_amount_mining": "200000",
                "total_amount_others": "200000.00",
                "total_amount_costs": "-400000.00",
            }
        ],
        "total": {
            "total_amount": "300000.00",
            "total_amount_ess": "1133333.33",
            "total_amount_all": "1433333.33",
            "total_amount_mining": "200000.00",
            "total_amount_others": "200000.00",
            "total_amount_costs": "-400000.00",
        },
    }
]

CharNoMining = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "entity_type": "character",
                "total_amount": "200000.00",
                "total_amount_ess": "1133333.33",
                "total_amount_mining": "0.00",
                "total_amount_others": "200000.00",
                "total_amount_costs": "-400000.00",
            }
        ],
        "total": {
            "total_amount": "200000.00",
            "total_amount_ess": "1133333.33",
            "total_amount_all": "1133333.33",
            "total_amount_mining": "0.00",
            "total_amount_others": "200000.00",
            "total_amount_costs": "-400000.00",
        },
    }
]



from datetime import datetime

from ninja import NinjaAPI

from ledger.api import schema
from ledger.api.api_helper.character_helper import CharacterProcess
from ledger.api.api_helper.corporation_helper import CorporationProcess
from ledger.api.helpers import (
    get_alliance,
    get_alts_queryset,
    get_character,
    get_corporation,
)
from ledger.hooks import get_extension_logger
from ledger.models import CorporationAudit

logger = get_extension_logger(__name__)


# pylint: disable=too-many-function-args
def ledger_api_process(request, entity_type: str, entity_id: int, date: str, view: str):
    request_main = request.GET.get("main", False)
    perm = True

    if entity_type == "corporation":
        perm, entitys = get_corporation(request, entity_id)
    elif entity_type == "character":
        perm, entitys = get_character(request, entity_id)
    elif entity_type == "alliance":
        perm, entitys = get_alliance(request, entity_id)
        # Get all corporations in the alliance
        if entitys:
            entitys = CorporationAudit.objects.filter(
                corporation__alliance__alliance_id__in=entitys
            ).values_list("corporation__corporation_id", flat=True)

    if perm is False:
        return "Permission Denied", None

    if perm is None:
        return None, None

    if entity_type == "character":
        if entity_id == 0 or request_main:
            characters = get_alts_queryset(entitys)
        else:
            characters = [entitys]
        return CharacterProcess(characters, date, view), entitys

    if entity_type == "corporation":
        return CorporationProcess(entitys, date, view), entitys

    if entity_type == "alliance":
        return CorporationProcess(entitys, date, view), entitys

    return "Wrong Entity Type", None


class LedgerApiEndpoints:
    tags = ["Ledger"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "{entity_type}/{entity_id}/ledger/date/{date}/view/{view}/",
            response={200: list[schema.Ledger], 403: str, 404: str},
            tags=self.tags,
        )
        def get_ledger(request, entity_type: str, entity_id: int, date: str, view: str):
            try:
                date_obj = datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                return 403, "Invalid Date format. Use YYYY-MM-DD"

            ledger, _ = ledger_api_process(
                request, entity_type, entity_id, date_obj, view
            )

            if isinstance(ledger, str):
                return 403, ledger

            if ledger is None:
                return 404, "No data found"

            output = ledger.generate_ledger()
            return output

        @api.get(
            "{entity_type}/{entity_id}/billboard/date/{date}/view/{view}/",
            response={200: list[schema.Billboard], 403: str, 404: str},
            tags=self.tags,
        )
        def get_billboard_ledger(
            request, entity_type: str, entity_id: int, date: str, view: str
        ):
            try:
                date_obj = datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                return 403, "Invalid Date format. Use YYYY/MM/DD"

            ledger, entitys = ledger_api_process(
                request, entity_type, entity_id, date_obj, view
            )

            if isinstance(ledger, str):
                return 403, ledger

            if ledger is None:
                return 404, "No data found"

            output = ledger.generate_billboard(entitys)
            return output



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext as trans

from ledger.api.helpers import get_alts_queryset
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import CharacterWalletJournalEntry
from ledger.models.corporationaudit import CorporationWalletJournalEntry
from ledger.view_helpers.core import calculate_ess_stolen

logger = get_extension_logger(__name__)


@dataclass
class TemplateData:
    """TemplateData class to store the data."""

    request: any
    main: any
    date: datetime
    view: str
    corporations_ids: list = None
    current_date: timezone.datetime = None

    def __post_init__(self):
        self.ledger_date = self.current_date
        self.ledger_date = self.ledger_date.replace(year=self.date.year)
        if self.view == "month":
            self.ledger_date = self.ledger_date.replace(month=self.date.month)
        elif self.view == "day":
            self.ledger_date = self.ledger_date.replace(
                month=self.date.month, day=self.date.day
            )


class TemplateProcess:
    """TemplateProcess class to process the data."""

    def __init__(self, chars: list, data: TemplateData, show_year=False):
        self.data = data
        self.chars = chars
        self.show_year = show_year
        self.template_dict = {}

    def filter_date(self):
        """Filter the date."""
        filter_date = Q(date__year=self.data.date.year)
        if self.data.view == "month":
            filter_date &= Q(date__month=self.data.date.month)
        elif self.data.view == "day":
            filter_date &= Q(date__month=self.data.date.month)
            filter_date &= Q(date__day=self.data.date.day)
        return filter_date

    # pylint: disable=duplicate-code
    def character_template(self):
        """
        Create the character template.
        return: dict
        """
        # Create Char List
        chars = self.chars
        chars_ids = [char.character_id for char in self.chars]

        filter_date = self.filter_date()

        character_journal, mining_journal, corporation_journal = (
            CharacterWalletJournalEntry.objects.filter(
                filter_date,
            )
            .select_related("first_party", "second_party")
            .generate_ledger(
                characters=chars,
                filter_date=filter_date,
                exclude=chars_ids,
            )
        )
        mining_journal = mining_journal.annotate_pricing()

        self._process_characters(character_journal, corporation_journal, mining_journal)
        return self.template_dict

    def corporation_template(self):
        """
        Create the corporation template.
        return: dict
        """

        filter_date = self.filter_date()

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(
                filter_date,
            )
            .select_related("first_party", "second_party")
            .order_by("-date")
        )

        self._process_corporation(corporation_journal)
        return self.template_dict

    # Process the character
    def _process_characters(
        self, character_journal, corporation_journal, mining_journal
    ):
        """Process the characters."""
        # Get the alts of the main character
        alts = get_alts_queryset(self.data.main)
        chars_list = [char.character_id for char in alts]

        # Get journals
        models = character_journal, corporation_journal, mining_journal
        # Process the amounts
        amounts = self._process_amounts_char(models, chars_list)

        main_name = (
            self.data.main.character_name if not self.show_year else trans("Summary")
        )
        main_id = self.data.main.character_id if not self.show_year else 0

        self._update_template_dict(main_id, main_name)
        self._generate_amounts_dict(amounts)

    # Process the corporation
    def _process_corporation(self, corporation_journal):
        """Process the corporations."""
        chars = [char.eve_id for char in self.chars]
        # Process the amounts
        amounts = defaultdict(lambda: defaultdict(Decimal))

        amounts = corporation_journal.generate_template(
            amounts=amounts,
            character_ids=chars,
            corporations_ids=self.data.corporations_ids,
            filter_date=self.data.ledger_date,
            entity_type="corporation",
        )

        amounts["stolen"] = defaultdict(Decimal)
        amounts = calculate_ess_stolen(amounts)

        main_name = (
            self.data.main.character_name if not self.show_year else trans("Summary")
        )
        main_id = self.data.main.character_id if not self.show_year else 0

        # Update the template dict
        self._update_template_dict(main_id, main_name)
        self._generate_amounts_dict(amounts)

    # Update Core Dict
    def _update_template_dict(self, main_id=0, main_name=trans("Unknown")):
        date = (
            str(self.data.ledger_date.year)
            if self.data.view == "year"
            else self.data.ledger_date.strftime("%B")
        )
        self.template_dict.update(
            {
                "main_name": main_name,
                "main_id": main_id,
                "date": date,
            }
        )

    # Add Amounts to Dict
    def _generate_amounts_dict(self, amounts):
        """Generate the amounts dictionary."""
        current_day = 365 if self.data.view == "year" else self.data.ledger_date.day
        # Calculate the total sum
        total_sum = sum(
            amounts[key]["total_amount"] for key in amounts if key != "stolen"
        )

        total_current_day_sum = sum(
            amounts[key]["total_amount_day"] for key in amounts if key != "stolen"
        )

        self.template_dict.update(
            {
                key: {
                    sub_key: value
                    for sub_key, value in {
                        "total_amount": amounts[key]["total_amount"],
                        "total_amount_day": (
                            amounts[key]["total_amount_day"]
                            if self.data.view == "month"
                            and self.data.current_date.month == self.data.date.month
                            else 0  # Only show daily amount if not year and in the correct month
                        ),
                        "total_amount_hour": (
                            amounts[key]["total_amount_hour"]
                            if key != "mining"
                            # Bad Fix but it works..
                            else amounts[key]["total_amount_day"]
                        ),
                        "average_day": amounts[key]["total_amount"] / current_day,
                        "average_hour": (amounts[key]["total_amount"] / current_day)
                        / 24,
                        "average_tick": ((amounts[key]["total_amount"]) / 20),
                        "current_day_tick": (amounts[key]["total_amount_day"] / 20),
                        "average_day_tick": (
                            amounts[key]["total_amount"] / current_day / 20
                        ),
                        "average_hour_tick": (
                            amounts[key]["total_amount"] / current_day
                        )
                        / 24
                        / 20,
                    }.items()
                    if value != 0
                }
                for key in amounts
            }
        )

        self.template_dict["summary"] = {
            "total_amount": total_sum,
            "total_amount_day": total_sum / current_day,
            "total_amount_hour": (total_sum / current_day) / 24,
            "total_current_day": (
                total_current_day_sum
                if self.data.view == "month"
                and self.data.current_date.month == self.data.date.month
                else 0
            ),  # Only show daily amount if not year and in the correct month
        }

    # Generate Amounts for all Chars
    # pylint: disable=too-many-locals
    def _process_amounts_char(self, models, chars_list):
        amounts = defaultdict(lambda: defaultdict(Decimal))

        # Set the models
        character_journal, corporation_journal, mining_journal = models

        # Generate Template for Character Journal
        amounts = character_journal.generate_template(
            amounts=amounts,
            character_ids=chars_list,
            filter_date=self.data.ledger_date,
            exclude=chars_list,
        )

        # Generate Template for Mining Journal
        amounts = mining_journal.generate_template(
            amounts=amounts,
            chars_list=chars_list,
            filter_date=self.data.ledger_date,
        )

        # Generate Template for Corporation Journal
        amounts = corporation_journal.generate_template(
            amounts=amounts,
            character_ids=chars_list,
            corporations_ids=self.data.corporations_ids,
            filter_date=self.data.ledger_date,
        )

        # Calculate the stolen ESS
        amounts["stolen"] = defaultdict(Decimal)
        amounts = calculate_ess_stolen(amounts)
        return amounts



import calendar
from dataclasses import asdict, dataclass, field
from datetime import datetime
from decimal import Decimal

from django.utils import timezone

from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


@dataclass
class LedgerDataCore:
    total_bounty: Decimal = Decimal("0.00")
    total_ess_payout: Decimal = Decimal("0.00")
    total_mining: Decimal = Decimal("0.00")
    total_miscellaneous: Decimal = Decimal("0.00")
    total_isk: Decimal = Decimal("0.00")


@dataclass
class LedgerData(LedgerDataCore):
    total_cost: Decimal = Decimal(0)
    total_production_cost: Decimal = Decimal(0)
    total_market_cost: Decimal = Decimal(0)


class LedgerModels:
    """class to store the models."""

    def __init__(
        self, character_journal=None, corporation_journal=None, mining_journal=None
    ):
        self.char_journal = character_journal
        self.corp_journal = corporation_journal
        self.mining_journal = mining_journal


@dataclass
class LedgerDate:
    """class to store the date."""

    year: int
    month: int
    monthly: bool = field(init=False)
    current_date: datetime = None
    range_data: int = field(init=False)
    day_checks: list = field(init=False)

    def calculate_days(self):
        _, num_days = calendar.monthrange(self.year, self.month)
        return num_days

    def __post_init__(self):
        self.current_date = timezone.now()
        self.monthly = self.month == 0
        self.range_data = 12 if self.monthly else self.calculate_days()
        self.day_checks = list(range(1, self.range_data + 1))


@dataclass
class LedgerTotal:
    """class to store the total amounts."""

    total_amount: Decimal = Decimal("0.00")
    total_amount_ess: Decimal = Decimal("0.00")
    total_amount_mining: Decimal = Decimal("0.00")
    total_amount_others: Decimal = Decimal("0.00")
    total_amount_costs: Decimal = Decimal("0.00")
    total_amount_all: Decimal = Decimal("0.00")

    def to_dict(self):
        return asdict(self)

    def get_summary(self, totals: dict):
        self.total_amount += totals.get("total_amount", 0)
        self.total_amount_ess += totals.get("total_amount_ess", 0)
        self.total_amount_all += totals.get("total_amount_all", 0)
        self.total_amount_mining += totals.get("total_amount_mining", 0)
        self.total_amount_others += totals.get("total_amount_others", 0)
        self.total_amount_costs += totals.get("total_amount_costs", 0)

    def calculate_total_sum(self, characters_dict: dict):
        for _, char_data in characters_dict.items():
            self.total_amount_all += (
                char_data["total_amount"]
                + char_data["total_amount_ess"]
                + char_data["total_amount_mining"]
                + char_data["total_amount_others"]
                - abs(char_data["total_amount_costs"])
            )


@dataclass
class LedgerCharacterDict(LedgerTotal):
    """class to store the character dictionary."""

    characters: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return self.characters

    def get_default_character_dict(self, char_id=0, char_name="Unknown") -> dict:
        """Return a default character dictionary."""
        return {
            "main_id": char_id,
            "main_name": char_name,
            "entity_type": "character",
            "total_amount": self.total_amount,
            "total_amount_ess": self.total_amount_ess,
            "total_amount_mining": self.total_amount_mining,
            "total_amount_others": self.total_amount_others,
            "total_amount_costs": self.total_amount_costs,
        }

    def add_or_update_character(self, char_id, char_name, **kwargs):
        """Add or update a character in the dictionary."""
        if char_id not in self.characters:
            self.characters[char_id] = self.get_default_character_dict(
                char_id, char_name
            )

        for key, value in kwargs.items():
            if key in self.characters[char_id]:
                self.characters[char_id][key] = value

    def add_amount_to_character(self, char_id, char_name, amount, key):
        """Add an amount to a character."""
        if char_id not in self.characters:
            self.characters[char_id] = self.get_default_character_dict(
                char_id, char_name
            )

        if char_id in self.characters:
            self.characters[char_id][key] += amount



from typing import TYPE_CHECKING

from django.db import models
from django.utils import timezone
from eveuniverse.models import EveType

from ledger.constants import COMMAND_CENTER, EXTRACTOR_CONTROL_UNIT, SPACEPORTS
from ledger.hooks import get_extension_logger

if TYPE_CHECKING:  # pragma: no cover
    from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails

logger = get_extension_logger(__name__)


class PlanetaryQuerySet(models.QuerySet):
    def get_or_create_facilitys(self, planet: "CharacterPlanet"):
        """Get or Create Facilitys for a given Planet"""
        return self._get_or_create_facilitys(planet=planet)

    def _get_or_create_facilitys(self, planet: "CharacterPlanet"):
        try:
            facilitys = self.get(planet=planet)
            facility = facilitys.facilitys
            created = False
            return facility, created
        except self.model.DoesNotExist:
            facility, created = self.update_or_create_facilitys(planet=planet)
        return facility, created

    def update_or_create_layout(
        self, planet: "CharacterPlanetDetails", esi_result: dict
    ):
        """Update or Create Layout for a given Planet"""
        return self._update_or_create(planet=planet, esi_result=esi_result)

    def _update_or_create(self, planet: "CharacterPlanetDetails", esi_result: dict):
        planetdetails, created = self.update_or_create(
            planet=planet,
            defaults={
                "links": esi_result["links"],
                "pins": esi_result["pins"],
                "routes": esi_result["routes"],
                "facilitys": None,
                "last_update": timezone.now(),
            },
        )

        self._update_facility(planetdetails)

        logger.debug("Planet %s Facilitys Updated", planetdetails)

        return planetdetails, created

    def _update_facility(self, planet: "CharacterPlanetDetails"):
        facility_info = self.get_facility_info(planet)
        logger.debug("Facility Info: %s", facility_info)
        planet.facilitys = facility_info
        planet.save()
        logger.debug("Planet %s saved with updated facilitys", planet)
        logger.debug("Facilitys: %s", planet.facilitys)
        return planet

    # TODO make code easier or split it to peaces for better readable
    def get_facility_info(self, planet: "CharacterPlanetDetails"):
        facility_info = {}

        # Process pins
        for pin in planet.pins:
            pin_id = pin["pin_id"]
            item_type, _ = EveType.objects.get_or_create_esi(id=pin["type_id"])
            if (
                pin["type_id"] in SPACEPORTS
                or pin["type_id"] in EXTRACTOR_CONTROL_UNIT
                or pin["type_id"] in COMMAND_CENTER
            ):
                continue
            facility_info[pin_id] = {
                "facility_id": item_type.id,
                "facility_name": item_type.name,
                "resources": [],
                "storage": {
                    content["type_id"]: content["amount"]
                    for content in pin.get("contents", [])
                },
            }

        self._facility_production_chain(planet, facility_info)

        self._facility_produce_depend(planet, facility_info)

        return facility_info

    def _facility_production_chain(
        self, planet: "CharacterPlanetDetails", facility_info: dict
    ):
        """Update dict and add all Production Information to the given dict"""

        extractors = planet.get_extractors_info()
        extractors_item_ids = [
            extractor["item_id"] for extractor in extractors.values()
        ]

        for route in planet.routes:
            destination_pin_id = route["destination_pin_id"]
            source_pin_id = route["source_pin_id"]
            content_type, _ = EveType.objects.get_or_create_esi(
                id=route["content_type_id"]
            )

            if destination_pin_id in facility_info:
                req_quantity = route["quantity"]
                current_quantity = facility_info[destination_pin_id]["storage"].get(
                    content_type.id, 0
                )
                missing_quantity = req_quantity - current_quantity
                still_producing = (
                    content_type.id in extractors_item_ids
                    if not planet.is_expired
                    else False
                )

                resource = {
                    "item_id": content_type.id,
                    "item_name": content_type.name,
                    "req_quantity": req_quantity,
                    "current_quantity": current_quantity,
                    "missing_quantity": max(missing_quantity, 0),
                    "still_producing": still_producing,
                }
                facility_info[destination_pin_id]["resources"].append(resource)

            if source_pin_id in facility_info:
                facility_info[source_pin_id]["output_product"] = {
                    "item_id": content_type.id,
                    "item_name": content_type.name,
                    "output_quantity": route["quantity"],
                }
        return facility_info

    def _facility_produce_depend(self, planet, facility_info):
        """Check if Facility is Producing"""
        for facility in facility_info.values():
            for resource in facility["resources"]:
                self._facility_still_producing(planet, resource, facility_info)
        return facility_info

    def _facility_still_producing(self, planet, resource, facility_info) -> bool:
        """Check if ressource exist in Production Chain and change State"""
        for other_facility in facility_info.values():
            if (
                "output_product" in other_facility
                and other_facility["output_product"]["item_id"] == resource["item_id"]
                and not planet.is_expired
            ):
                if other_facility["output_product"]["output_quantity"] > 0:
                    resource["still_producing"] = True


class PlanetaryManagerBase(models.Manager):
    pass


PlanetaryManager = PlanetaryManagerBase.from_queryset(PlanetaryQuerySet)



"""Generate AllianceAuth test objects from allianceauth.json."""

import json
from datetime import date, datetime
from pathlib import Path

from eveuniverse.models import EveMarketPrice, EvePlanet, EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from ledger.models.characteraudit import CharacterAudit
from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails
from ledger.tests.testdata.planetary import planetary_data


def load_planetary():
    """Load test data for planetary."""
    CharacterPlanet.objects.all().delete()
    CharacterPlanet.objects.update_or_create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        planet=EvePlanet.objects.get(id=4001),
        defaults={
            "upgrade_level": 5,
            "num_pins": 5,
            "last_update": None,
        },
    )
    CharacterPlanet.objects.update_or_create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        planet=EvePlanet.objects.get(id=4002),
        defaults={
            "upgrade_level": 5,
            "num_pins": 5,
            "last_update": None,
        },
    )

    CharacterPlanet.objects.update_or_create(
        character=CharacterAudit.objects.get(character__character_id=1002),
        planet=EvePlanet.objects.get(id=4001),
        defaults={
            "upgrade_level": 5,
            "num_pins": 5,
            "last_update": None,
        },
    )

    CharacterPlanetDetails.objects.all().delete()
    CharacterPlanetDetails.objects.update_or_create(
        planet=CharacterPlanet.objects.get(
            planet__id=4001, character__character__character_name="Gneuten"
        ),
        defaults=planetary_data,
    )
    CharacterPlanetDetails.objects.update_or_create(
        planet=CharacterPlanet.objects.get(
            planet__id=4002, character__character__character_name="Gneuten"
        ),
        defaults=planetary_data,
    )

    CharacterPlanetDetails.objects.update_or_create(
        planet=CharacterPlanet.objects.get(
            planet__id=4001, character__character__character_id=1002
        ),
        defaults=planetary_data,
    )



noData = [
    {
        "billboard": {
            "standard": {
                "charts": None,
                "rattingbar": None,
                "workflowgauge": None,
            },
        },
    }
]

CharmonthlyMarch = [
    {
        "billboard": {
            "standard": {
                "charts": {
                    "title": "Donut Chart",
                    "date": "2025-01-31",
                    "categories": [
                        "bounty",
                        "contract",
                        "donation",
                        "ess",
                        "insurance",
                        "market",
                        "mining",
                        "production",
                    ],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY (income)": {"value": 200000.00, "mode": "income"},
                            "CONTRACT (income)": {"value": 100000.00, "mode": "income"},
                            "DONATION (income)": {"value": 100000.00, "mode": "income"},
                            "ESS (income)": {"value": 1133333.33, "mode": "income"},
                            "INSURANCE (cost)": {"value": 100000.00, "mode": "cost"},
                            "MARKET (cost)": {"value": 200000.00, "mode": "cost"},
                            "MINING (income)": {"value": 200000.00, "mode": "income"},
                            "PRODUCTION (cost)": {"value": 100000.00, "mode": "cost"},
                        }
                    ],
                },
                "rattingbar": {
                    "title": "Ratting Chart",
                    "date": "2025-01-31",
                    "categories": ["Bounty", "ESS", "Miscellaneous", "Mining"],
                    "series": [
                        {"date": "2024-03-16", "mining": 200000},
                        {
                            "date": "2024-03-19",
                            "bounty": 200000,
                            "ess": 1133333,
                            "miscellaneous": 200000,
                        },
                    ],
                },
                "workflowgauge": {
                    "title": "Workflow Chart",
                    "date": "2025-01-31",
                    "categories": ["bounty", "ess", "mining", "miscellaneous"],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY": {"value": 200000.00, "mode": "income"},
                            "ESS": {"value": 1133333.33, "mode": "income"},
                            "MINING": {"value": 200000.00, "mode": "income"},
                            "MISCELLANEOUS": {"value": 200000.00, "mode": "income"},
                        }
                    ],
                },
            }
        }
    }
]


CharmonthlyMarchMulti = [
    {
        "billboard": {
            "standard": {
                "charts": {
                    "title": "Donut Chart",
                    "date": "2025-01-31",
                    "categories": [
                        "bounty",
                        "contract",
                        "donation",
                        "ess",
                        "insurance",
                        "market",
                        "mining",
                        "production",
                    ],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY (income)": {"value": 200000.00, "mode": "income"},
                            "CONTRACT (income)": {"value": 100000.00, "mode": "income"},
                            "DONATION (income)": {"value": 100000.00, "mode": "income"},
                            "ESS (income)": {"value": 1133333.33, "mode": "income"},
                            "INSURANCE (cost)": {"value": 100000.00, "mode": "cost"},
                            "MARKET (cost)": {"value": 200000.00, "mode": "cost"},
                            "MINING (income)": {"value": 300000.00, "mode": "income"},
                            "PRODUCTION (cost)": {"value": 100000.00, "mode": "cost"},
                        }
                    ],
                },
                "rattingbar": {
                    "title": "Ratting Chart",
                    "date": "2025-01-31",
                    "categories": ["Bounty", "ESS", "Miscellaneous", "Mining"],
                    "series": [
                        {"date": "2024-03-16", "mining": 300000},
                        {
                            "date": "2024-03-19",
                            "bounty": 200000,
                            "ess": 1133333,
                            "miscellaneous": 200000,
                        },
                    ],
                },
                "workflowgauge": {
                    "title": "Workflow Chart",
                    "date": "2025-01-31",
                    "categories": ["bounty", "ess", "mining", "miscellaneous"],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY": {"value": 200000.00, "mode": "income"},
                            "ESS": {"value": 1133333.33, "mode": "income"},
                            "MINING": {"value": 300000.00, "mode": "income"},
                            "MISCELLANEOUS": {"value": 200000.00, "mode": "income"},
                        }
                    ],
                },
            }
        }
    }
]

Charyearly = [
    {
        "billboard": {
            "standard": {
                "charts": {
                    "title": "Donut Chart",
                    "date": "2025-01-31",
                    "categories": [
                        "bounty",
                        "contract",
                        "donation",
                        "ess",
                        "insurance",
                        "market",
                        "mining",
                        "production",
                    ],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY (income)": {"value": 300000.00, "mode": "income"},
                            "CONTRACT (income)": {"value": 100000.00, "mode": "income"},
                            "DONATION (income)": {"value": 100000.00, "mode": "income"},
                            "ESS (income)": {"value": 1133333.33, "mode": "income"},
                            "INSURANCE (cost)": {"value": 100000.00, "mode": "cost"},
                            "MARKET (cost)": {"value": 200000.00, "mode": "cost"},
                            "MINING (income)": {"value": 200000.00, "mode": "income"},
                            "PRODUCTION (cost)": {"value": 100000.00, "mode": "cost"},
                        }
                    ],
                },
                "rattingbar": {
                    "title": "Ratting Chart",
                    "date": "2025-01-31",
                    "categories": ["Bounty", "ESS", "Miscellaneous", "Mining"],
                    "series": [
                        {"date": "2024-01", "bounty": 100000},
                        {
                            "date": "2024-03",
                            "bounty": 200000,
                            "ess": 1133333,
                            "miscellaneous": 200000,
                            "mining": 200000,
                        },
                    ],
                },
                "workflowgauge": {
                    "title": "Workflow Chart",
                    "date": "2025-01-31",
                    "categories": [
                        "bounty",
                        "ess",
                        "mining",
                        "miscellaneous",
                    ],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY": {"value": 300000.00, "mode": "income"},
                            "ESS": {"value": 1133333.33, "mode": "income"},
                            "MINING": {"value": 200000.00, "mode": "income"},
                            "MISCELLANEOUS": {"value": 200000.00, "mode": "income"},
                        }
                    ],
                },
            }
        }
    }
]

CharNoMining = [
    {
        "billboard": {
            "standard": {
                "charts": {
                    "title": "Donut Chart",
                    "date": "2025-01-31",
                    "categories": [
                        "bounty",
                        "contract",
                        "donation",
                        "ess",
                        "insurance",
                        "market",
                        "production",
                    ],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY (income)": {"value": 200000.00, "mode": "income"},
                            "CONTRACT (income)": {"value": 100000.00, "mode": "income"},
                            "DONATION (income)": {"value": 100000.00, "mode": "income"},
                            "ESS (income)": {"value": 1133333.33, "mode": "income"},
                            "INSURANCE (cost)": {"value": 100000.00, "mode": "cost"},
                            "MARKET (cost)": {"value": 200000.00, "mode": "cost"},
                            "PRODUCTION (cost)": {"value": 100000.00, "mode": "cost"},
                        }
                    ],
                },
                "rattingbar": {
                    "title": "Ratting Chart",
                    "date": "2025-01-31",
                    "categories": ["Bounty", "ESS", "Miscellaneous", "Mining"],
                    "series": [
                        {
                            "date": "2024-03-19",
                            "bounty": 200000,
                            "ess": 1133333,
                            "miscellaneous": 200000,
                        }
                    ],
                },
                "workflowgauge": {
                    "title": "Workflow Chart",
                    "date": "2025-01-31",
                    "categories": ["bounty", "ess", "mining", "miscellaneous"],
                    "series": [
                        {
                            "date": "2025-01-31",
                            "BOUNTY": {"value": 200000.00, "mode": "income"},
                            "ESS": {"value": 1133333.33, "mode": "income"},
                            "MISCELLANEOUS": {"value": 200000.00, "mode": "income"},
                        }
                    ],
                },
            }
        }
    }
]



from dataclasses import dataclass, field
from typing import Any

from django.db.models import F
from django.db.models.functions import TruncDay, TruncHour, TruncMonth, TruncYear
from django.utils import timezone
from django.utils.translation import gettext as _

from ledger.api.api_helper.core_manager import LedgerModels
from ledger.api.helpers import convert_corp_tax
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


@dataclass
class _BillboardDict:
    charts: Any = field(default_factory=lambda: None)
    rattingbar: Any = field(default_factory=lambda: None)
    workflowgauge: Any = field(default_factory=lambda: None)


@dataclass
class BillboardDict:
    standard: _BillboardDict = field(default_factory=_BillboardDict)


@dataclass
class BillboardTrunc:
    year: tuple = field(default_factory=lambda: (TruncYear("date"), "%Y-%m"))
    month: tuple = field(default_factory=lambda: (TruncMonth("date"), "%Y-%m"))
    week: tuple = field(default_factory=lambda: (TruncDay("date"), "%Y-%m-%d"))
    day: tuple = field(default_factory=lambda: (TruncDay("date"), "%Y-%m-%d"))
    hour: tuple = field(
        default_factory=lambda: (TruncHour("date"), "%Y-%m-%d %H:00:00")
    )


@dataclass
class ChartData:
    title: str
    date: str
    categories: list[str]
    series: list[dict[str, Any]]


@dataclass
class BillboardSystem:
    data_points: dict = field(default_factory=dict)
    chord_data_points: dict = field(default_factory=dict)

    def _clean_category(self, category: str) -> str:
        """Remove the suffix from the category name."""
        return category.split("_")[0].lower()

    def add_or_update_data_point(
        self, date: str, category: str, value: float, data: dict = None
    ):
        mode = "cost" if "cost" in category.lower() else "income"
        display_category = self._clean_category(category)

        # Skip zero values
        if value == 0.0:
            return

        # Create a new data point if it does not exist
        if date not in self.data_points:
            self.data_points[date] = {}
        if display_category not in self.data_points[date]:
            self.data_points[date][display_category] = []

        # Add the new data point
        self.data_points[date][display_category].append(
            {
                "value": value,
                "mode": mode,
                "data": data,
            }
        )

    def add_chord_data_point(self, from_char: str, to: str, value: float):
        if value == 0.0:
            return

        if from_char not in self.chord_data_points:
            self.chord_data_points[from_char] = []

        # Check if from char has already a value for the 'to' value
        for data_point in self.chord_data_points[from_char]:
            if data_point["to"] == to:
                data_point["value"] += value
                break
        else:
            # Add a new data point
            self.chord_data_points[from_char].append(
                {"from": from_char, "to": to, "value": value}
            )

    def to_xy(self, title: str, is_character=False) -> ChartData:
        """Convert the data points to a XY chart."""
        # Define the specific categories we are interested in
        categories = {
            True: (
                ["bounty", "ess", "miscellaneous"],
                ["Bounty", "ESS", "Miscellaneous"],
            ),
            False: (
                ["bounty", "ess", "miscellaneous", "mining"],
                ["Bounty", "ESS", "Miscellaneous", "Mining"],
            ),
        }

        specific_categories, category_names = categories[is_character]

        # Create series data
        series = []
        for date, categories in self.data_points.items():
            data_entry = {"date": date}
            for category in specific_categories:
                total_value = sum(
                    data["value"] for data in categories.get(category, [])
                )
                if total_value != 0.0:
                    data_entry[category] = round(total_value)
            series.append(data_entry)

        # Sort series data by date
        series.sort(key=lambda x: x["date"])

        date = timezone.now().strftime("%Y-%m-%d")
        return ChartData(
            title=title, date=date, categories=category_names, series=series
        )

    def _to_chart(
        self, title: str, included_categories: set, include_mode=False
    ) -> ChartData:
        """Convert the data points to a chart."""
        data_entry = {"date": timezone.now().strftime("%Y-%m-%d")}
        # Initialize dictionaries to hold the aggregated values for cost and income
        aggregated_cost_data = {
            category: {"value": 0.00, "mode": None} for category in included_categories
        }
        aggregated_income_data = {
            category: {"value": 0.00, "mode": None} for category in included_categories
        }

        # Aggregate the values for each category across all dates
        for date, categories in self.data_points.items():
            for category, data_list in categories.items():
                if category in included_categories:
                    for data in data_list:
                        if data["mode"] == "cost":
                            aggregated_cost_data[category]["value"] += abs(
                                data["value"]
                            )
                            aggregated_cost_data[category]["mode"] = data["mode"]
                        elif data["mode"] == "income":
                            aggregated_income_data[category]["value"] += abs(
                                data["value"]
                            )
                            aggregated_income_data[category]["mode"] = data["mode"]

        # Add the aggregated values to the output
        for category, value in aggregated_cost_data.items():
            if value["value"] == 0.0:
                continue
            display_category = f"{category.upper()}" + (
                f" ({value['mode']})" if include_mode else ""
            )
            data_entry[display_category] = {
                "value": float(f"{value['value']:.2f}"),
                "mode": value["mode"],
            }

        for category, value in aggregated_income_data.items():
            if value["value"] == 0.0:
                continue
            display_category = f"{category.upper()}" + (
                f" ({value['mode']})" if include_mode else ""
            )
            data_entry[display_category] = {
                "value": float(f"{value['value']:.2f}"),
                "mode": value["mode"],
            }

        # Create series data with a single entry
        series = [data_entry]
        categories = sorted(list(included_categories))

        date = timezone.now().strftime("%Y-%m-%d")
        return ChartData(title=title, date=date, categories=categories, series=series)

    def _to_chord_chart(self, title: str) -> ChartData:
        """Convert the data points to a bubble chart."""
        series = []
        max_chords = []
        others_summary = {}

        # Fügen Sie die aggregierten Werte zur Ausgabe hinzu
        for main_char, summaries in self.chord_data_points.items():
            if len(max_chords) <= 15:
                max_chords.append(main_char)
            for summary in summaries:
                if main_char in max_chords:
                    series.append(
                        {
                            "from": summary["from"],
                            "to": summary["to"],
                            "value": summary["value"],
                            "main": summary["from"],
                        }
                    )
                else:
                    to = summary["to"]
                    if to not in others_summary:
                        others_summary[to] = 0
                    others_summary[to] += summary["value"]

        # Add summarized "Others" entries to output
        for to_corp, values in others_summary.items():
            series.append(
                {
                    "from": "Others",
                    "to": to_corp,
                    "value": values,
                    "main": "Others",
                }
            )

        date = timezone.now().strftime("%Y-%m-%d")
        return ChartData(
            title=title,
            date=date,
            categories=None,
            series=series,
        )

    def to_xy_data(self, is_character: bool) -> ChartData:
        return self.to_xy(title="Ratting Chart", is_character=is_character)

    def to_chord_data(self) -> ChartData:
        return self._to_chord_chart(title="Chord Chart")

    def to_chart_data(self) -> ChartData:
        excluded_categories = {"costs", "miscellaneous"}
        all_categories = set()

        # Collect all unique categories from the data points
        for categories in self.data_points.values():
            all_categories.update(categories.keys())

        included_categories = all_categories - excluded_categories
        return self._to_chart(
            title="Donut Chart",
            included_categories=included_categories,
            include_mode=True,
        )

    def to_gauge_data(self) -> ChartData:
        included_categories = {"bounty", "ess", "mining", "miscellaneous"}
        return self._to_chart(
            title="Workflow Chart", included_categories=included_categories
        )


class BillboardLedger:
    def __init__(self, view: str, models: LedgerModels, corp=False):
        self.view = view
        self.models = models
        self.is_corp = corp
        self.billboard_dict = BillboardDict()

    def _sort_series(self, chart_data: ChartData) -> ChartData:
        """Sort the series data by category names."""
        for series_item in chart_data.series:
            sorted_series = {
                k: v for k, v in sorted(series_item.items()) if k != "date"
            }
            sorted_series = {"date": series_item["date"], **sorted_series}
            series_item.clear()
            series_item.update(sorted_series)
        return chart_data

    def _process_corp_chart(self, billboard: BillboardSystem, corp_qs, period_format):
        """Process Corporation Chart from Corporation Journal."""
        for entry in corp_qs:
            date = entry["period"].strftime(period_format)
            main_char_name = entry.get("main_entity_name", "Unknown")
            corporation = entry.get("corporation", "Unknown")
            bounty = entry.get("bounty_income", 0)
            ess = entry.get("ess_income", 0)
            miscellaneous = entry.get("miscellaneous", 0)
            values = bounty + ess + miscellaneous

            for key, value in entry.items():
                if key in [
                    "period",
                    "alts",
                    "main_entity_name",
                    "main_entity_id",
                    "corporation",
                ]:
                    continue
                billboard.add_or_update_data_point(
                    date=date, category=key, value=float(value)
                )

            billboard.add_chord_data_point(
                from_char=main_char_name, to=corporation, value=values
            )

    def _process_char_chart(self, billboard: BillboardSystem, corp_qs, period_format):
        """Process Character Chart from Corporation Journal."""
        for entry in corp_qs:
            date = entry["period"].strftime(period_format)
            for key, value in entry.items():
                if key in [
                    "period",
                    "alts",
                ]:
                    continue
                if not self.is_corp and key in ["bounty_income"]:
                    continue
                value = convert_corp_tax(value)
                billboard.add_or_update_data_point(
                    date=date, category=key, value=float(value)
                )

    # pylint: disable=too-many-branches, too-many-locals
    def _process_billboard(
        self, billboard: BillboardSystem, annotations, period_format
    ) -> BillboardSystem:
        """Process the queryset for the billboard."""
        corp_qs = self.models.corp_journal.annotate(**annotations)

        if self.is_corp:
            corp_qs = corp_qs.values(
                "period", "main_entity_id", "main_entity_name"
            ).annotate(
                corporation=F("division__corporation__corporation__corporation_name")
            )
        else:
            corp_qs = corp_qs.values("period")

        corp_qs = (
            corp_qs.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
            .annotate_daily_goal_income()
        )

        # Get Character Data
        if not self.is_corp:
            char_qs = (
                self.models.char_journal.annotate(**annotations)
                .values("period")
                .annotate_billboard(self.chars, self.alts)
            )

            mining_qs = (
                self.models.mining_journal.annotate(**annotations)
                .values("period")
                .annotate_billboard(self.chars)
            )

            # Process Character Journal
            for entry in char_qs:
                date = entry["period"].strftime(period_format)
                for key, value in entry.items():
                    if key not in ["period"]:
                        billboard.add_or_update_data_point(
                            date=date, category=key, value=float(value)
                        )
            # Process Mining Journal
            for entry in mining_qs:
                date = entry["period"].strftime(period_format)
                for key, value in entry.items():
                    if key not in ["period"]:
                        billboard.add_or_update_data_point(
                            date=date, category="mining", value=float(value)
                        )
            # Process Corp Journal
            self._process_char_chart(billboard, corp_qs, period_format)
        if self.is_corp:
            self._process_corp_chart(billboard, corp_qs, period_format)

        if not billboard.data_points:
            return None

        return billboard

    def annotate_days(self, period, billboard_dict: _BillboardDict, tick=False):
        """Generate the Billboard Data."""
        trunctype, period_format = period
        annotations = {"period": trunctype}
        self.tick = tick

        billboard = BillboardSystem()
        billboard = self._process_billboard(billboard, annotations, period_format)

        if billboard:
            # Create the Chart
            if self.is_corp:
                chart = billboard.to_chord_data()
            else:
                chart = billboard.to_chart_data()
                chart = self._sort_series(chart)
            billboard_dict.charts = chart

            # Create the Ratting Bar
            rattingbar = billboard.to_xy_data(is_character=self.is_corp)
            billboard_dict.rattingbar = rattingbar

            # Create the Gauge
            gauge = billboard.to_gauge_data()
            billboard_dict.workflowgauge = self._sort_series(gauge)
        return billboard_dict

    # Create the Billboard
    def billboard_ledger(self, chars, alts: list = None):
        """Generate the Billboard Ledger."""
        periods = BillboardTrunc()
        self.chars = chars
        self.alts = alts
        standard = None

        # Create Billboard Month
        if self.view == "year":
            standard = self.annotate_days(periods.month, self.billboard_dict.standard)
        elif self.view == "month":
            standard = self.annotate_days(periods.day, self.billboard_dict.standard)
        elif self.view == "day":
            standard = self.annotate_days(periods.hour, self.billboard_dict.standard)

        # Generate Billboard
        self.billboard_dict.standard = standard

        return self.billboard_dict



planet_many = [
    {
        "character_id": 1001,
        "character_name": "Gneuten",
        "planet": "Test Planet I",
        "planet_id": 4001,
        "planet_type_id": 13,
        "upgrade_level": 5,
        "expiry_date": "2024-08-26T17:17:02Z",
        "expired": True,
        "alarm": False,
        "extractors": {
            "1046238237375": {
                "install_time": "2024-08-12T17:17:02Z",
                "expiry_time": "2024-08-26T17:17:02Z",
                "item_id": 2268,
                "item_name": "Aqueous Liquids",
                "progress_percentage": 100.0,
            },
            "1046238237378": {
                "install_time": "2024-08-12T17:17:02Z",
                "expiry_time": "2024-08-26T17:17:02Z",
                "item_id": 2309,
                "item_name": "Ionic Solutions",
                "progress_percentage": 100.0,
            },
        },
        "products": {
            "raw": {
                "2268": {
                    "id": 2268,
                    "name": "Aqueous Liquids",
                    "category": "Planet Liquid-Gas - Raw Resource",
                },
                "2309": {
                    "id": 2309,
                    "name": "Ionic Solutions",
                    "category": "Planet Liquid-Gas - Raw Resource",
                },
            },
            "processed": {
                "9832": {
                    "id": 9832,
                    "name": "Coolant",
                    "category": "Refined Commodities - Tier 2",
                },
                "3645": {
                    "id": 3645,
                    "name": "Water",
                    "category": "Basic Commodities - Tier 1",
                },
                "2390": {"id": 2390, "name": "Electrolytes", "category": "Planet"},
            },
        },
        "storage": {
            "1036002868084": {
                "facility_id": 2557,
                "facility_name": "Storm Launchpad",
                "contents": [
                    {"amount": 14800, "type_id": 2390, "product_name": "Electrolytes"}
                ],
            },
            "1046238237381": {
                "facility_id": 2543,
                "facility_name": "Gas Launchpad",
                "contents": [],
            },
            "1046238237382": {
                "facility_id": 2543,
                "facility_name": "Gas Launchpad",
                "contents": [],
            },
        },
        "facility": None,
        "last_update": None,
    },
    {
        "character_id": 1001,
        "character_name": "Gneuten",
        "planet": "Test Planet I",
        "planet_id": 4002,
        "planet_type_id": 13,
        "upgrade_level": 5,
        "expiry_date": "2024-08-26T17:17:02Z",
        "expired": True,
        "alarm": False,
        "extractors": {
            "1046238237375": {
                "install_time": "2024-08-12T17:17:02Z",
                "expiry_time": "2024-08-26T17:17:02Z",
                "item_id": 2268,
                "item_name": "Aqueous Liquids",
                "progress_percentage": 100.0,
            },
            "1046238237378": {
                "install_time": "2024-08-12T17:17:02Z",
                "expiry_time": "2024-08-26T17:17:02Z",
                "item_id": 2309,
                "item_name": "Ionic Solutions",
                "progress_percentage": 100.0,
            },
        },
        "products": {
            "raw": {
                "2268": {
                    "id": 2268,
                    "name": "Aqueous Liquids",
                    "category": "Planet Liquid-Gas - Raw Resource",
                },
                "2309": {
                    "id": 2309,
                    "name": "Ionic Solutions",
                    "category": "Planet Liquid-Gas - Raw Resource",
                },
            },
            "processed": {
                "9832": {
                    "id": 9832,
                    "name": "Coolant",
                    "category": "Refined Commodities - Tier 2",
                },
                "3645": {
                    "id": 3645,
                    "name": "Water",
                    "category": "Basic Commodities - Tier 1",
                },
                "2390": {"id": 2390, "name": "Electrolytes", "category": "Planet"},
            },
        },
        "storage": {
            "1036002868084": {
                "facility_id": 2557,
                "facility_name": "Storm Launchpad",
                "contents": [
                    {"amount": 14800, "type_id": 2390, "product_name": "Electrolytes"}
                ],
            },
            "1046238237381": {
                "facility_id": 2543,
                "facility_name": "Gas Launchpad",
                "contents": [],
            },
            "1046238237382": {
                "facility_id": 2543,
                "facility_name": "Gas Launchpad",
                "contents": [],
            },
        },
        "facility": None,
        "last_update": None,
    },
]

planet_single = [
    {
        "character_id": 1001,
        "character_name": "Gneuten",
        "planet": "Test Planet I",
        "planet_id": 4001,
        "planet_type_id": 13,
        "upgrade_level": 5,
        "expiry_date": "2024-08-26T17:17:02Z",
        "expired": False,
        "alarm": False,
        "extractors": {
            "1046238237375": {
                "install_time": "2024-08-12T17:17:02Z",
                "expiry_time": "2024-08-26T17:17:02Z",
                "item_id": 2268,
                "item_name": "Aqueous Liquids",
                "progress_percentage": 57.14,
            },
            "1046238237378": {
                "install_time": "2024-08-12T17:17:02Z",
                "expiry_time": "2024-08-26T17:17:02Z",
                "item_id": 2309,
                "item_name": "Ionic Solutions",
                "progress_percentage": 57.14,
            },
        },
        "products": {
            "raw": {
                "2268": {
                    "id": 2268,
                    "name": "Aqueous Liquids",
                    "category": "Planet Liquid-Gas - Raw Resource",
                },
                "2309": {
                    "id": 2309,
                    "name": "Ionic Solutions",
                    "category": "Planet Liquid-Gas - Raw Resource",
                },
            },
            "processed": {
                "9832": {
                    "id": 9832,
                    "name": "Coolant",
                    "category": "Refined Commodities - Tier 2",
                },
                "3645": {
                    "id": 3645,
                    "name": "Water",
                    "category": "Basic Commodities - Tier 1",
                },
                "2390": {"id": 2390, "name": "Electrolytes", "category": "Planet"},
            },
        },
        "storage": {
            "1036002868084": {
                "facility_id": 2557,
                "facility_name": "Storm Launchpad",
                "contents": [
                    {"amount": 14800, "type_id": 2390, "product_name": "Electrolytes"}
                ],
            },
            "1046238237381": {
                "facility_id": 2543,
                "facility_name": "Gas Launchpad",
                "contents": [],
            },
            "1046238237382": {
                "facility_id": 2543,
                "facility_name": "Gas Launchpad",
                "contents": [],
            },
        },
        "facility": None,
        "last_update": None,
    }
]



from datetime import datetime

from django.db.models import Q

from ledger.api.api_helper.billboard_helper import BillboardLedger
from ledger.api.api_helper.core_manager import (
    LedgerCharacterDict,
    LedgerModels,
    LedgerTotal,
)
from ledger.api.helpers import get_alts_queryset
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import CharacterWalletJournalEntry

logger = get_extension_logger(__name__)


class CharacterProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(self, chars, date: datetime, view=None):
        self.date = date
        self.chars = chars
        self.view = view
        self.alts = self.get_alts(chars)
        self.chars_list = []
        self.summary_total = LedgerTotal()
        self.create_queryset()

    def create_queryset(self):
        """Create the Queryset for the Model"""
        # Get the All Alt Characters from Main
        if self.alts:
            self.chars_list = [char.character_id for char in self.alts]
        else:
            self.chars_list = [char.character_id for char in self.chars]

        # pylint: disable=duplicate-code
        filter_date = Q(date__year=self.date.year)
        if self.view == "month":
            filter_date &= Q(date__month=self.date.month)
        elif self.view == "day":
            filter_date &= Q(date__month=self.date.month)
            filter_date &= Q(date__day=self.date.day)

        # Process Ledger Models
        self.char_journal, self.mining_journal, self.corp_journal = (
            CharacterWalletJournalEntry.objects.filter(
                filter_date,
            )
            .select_related("first_party", "second_party")
            .generate_ledger(
                characters=self.chars,
                filter_date=filter_date,
                exclude=self.chars_list,
            )
        )

    def get_alts(self, main):
        """Get the Alts for the Main Character"""
        alts = None
        if len(main) == 1:
            # pylint: disable=broad-exception-caught
            try:
                alts = get_alts_queryset(main[0])
            except Exception:
                pass
        return alts

    # pylint: disable=too-many-locals
    def process_character_chars(self):
        """Process the characters for the Journal"""
        # Initialize character_dict with default values
        character_dict = LedgerCharacterDict()
        character_totals = LedgerTotal()

        # Annotate Mining for Characters
        mining_char_journal = self.mining_journal.annotate_mining()

        # Annotate Corp Journal for Characters
        corp_character_journal = (
            self.corp_journal.values("second_party__eve_id", "second_party__name")
            .annotate_daily_goal_income(is_character_ledger=True)
            .annotate_ess_income(is_character_ledger=True)
        )

        character_journal = self.char_journal

        for mining_char in mining_char_journal:
            char_name = mining_char.get(
                "character__character__character_name", "Unknown"
            )
            char_id = mining_char.get("character__character__character_id", 0)
            total_amount_mining = mining_char.get("total_amount", 0)

            character_dict.add_or_update_character(
                char_id,
                char_name,
                total_amount_mining=total_amount_mining,
            )

            totals = {
                "total_amount_mining": total_amount_mining,
            }
            # Summary all
            character_totals.get_summary(totals)

        for char in character_journal:
            char_name = char.get("char_name", "Unknown")
            char_id = char.get("char_id", 0)

            total_bounty = char.get("bounty_income", 0)
            total_others = char.get("miscellaneous", 0)
            total_costs = char.get("costs", 0)

            if total_bounty or total_others or total_costs:
                character_dict.add_or_update_character(
                    char_id,
                    char_name,
                    total_amount=total_bounty,
                    total_amount_others=total_others,
                    total_amount_costs=total_costs,
                )

            totals = {
                "total_amount": total_bounty,
                "total_amount_others": total_others,
                "total_amount_costs": total_costs,
            }
            # Summary all
            character_totals.get_summary(totals)

        for corp_char in corp_character_journal:
            char_name = corp_char.get("second_party__name", "Unknown")
            char_id = corp_char.get("second_party__eve_id", 0)

            total_ess = corp_char.get("ess_income", 0)
            total_daily_goal = corp_char.get("daily_goal_income", 0)

            if total_daily_goal:
                character_dict.add_amount_to_character(
                    char_id,
                    char_name,
                    total_daily_goal,
                    "total_amount_others",
                )

            if total_ess:
                character_dict.add_amount_to_character(
                    char_id,
                    char_name,
                    total_ess,
                    "total_amount_ess",
                )

            totals = {
                "total_amount_ess": total_ess,
                "total_amount_others": total_daily_goal,
            }
            # Summary all
            character_totals.get_summary(totals)

        # Generate the total sum
        character_totals.calculate_total_sum(character_dict.to_dict())

        return character_dict.to_dict(), character_totals.to_dict()

    def generate_ledger(self):
        # Create the Dicts for each Character
        character_dict, character_totals = self.process_character_chars()

        output = []
        output.append(
            {
                "ratting": sorted(
                    list(character_dict.values()), key=lambda x: x["main_name"]
                ),
                "total": character_totals,
            }
        )

        return output

    # pylint: disable=unused-argument
    def generate_billboard(self, corporations=None):
        mining_journal = self.mining_journal.annotate_pricing()

        # Create Data for Billboard
        models = LedgerModels(
            character_journal=self.char_journal,
            corporation_journal=self.corp_journal,
            mining_journal=mining_journal,
        )

        # Create the Billboard for the Characters
        ledger = BillboardLedger(view=self.view, models=models, corp=False)
        billboard_dict = ledger.billboard_ledger(self.chars_list, self.chars_list)

        output = []
        output.append(
            {
                "billboard": billboard_dict,
            }
        )

        return output



from datetime import datetime

from ninja import NinjaAPI

from django.shortcuts import render
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter

from ledger.api import schema
from ledger.api.api_helper.template_helper import TemplateData, TemplateProcess
from ledger.api.helpers import (
    get_alliance,
    get_alts_queryset,
    get_character,
    get_corp_alts_queryset,
    get_corporation,
    get_journal_entitys,
    get_main_and_alts_ids_corporations,
)
from ledger.hooks import get_extension_logger
from ledger.models import CorporationAudit
from ledger.models.general import EveEntity

logger = get_extension_logger(__name__)


def ledger_api_process(request, entity_type: str, entity_id: int):
    perm = None
    if entity_type == "corporation":
        perm, entitys = get_corporation(request, entity_id)

    if entity_type == "alliance":
        perm, entitys = get_alliance(request, entity_id)
        # Get all corporations in the alliance
        entitys = CorporationAudit.objects.filter(
            corporation__alliance__alliance_id__in=entitys
        ).values_list("corporation__corporation_id", flat=True)

    return perm, entitys


class LedgerTemplateApiEndpoints:
    tags = ["LedgerInformationDetails"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "character/{character_id}/template/date/{date}/view/{view}/",
            response={200: list[schema.CharacterLedgerTemplate], 403: str},
            tags=self.tags,
        )
        def get_ledger_char_details_information(
            request, character_id: int, date: str, view: str
        ):
            try:
                date_obj = datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                return 403, "Invalid Date format. Use YYYY-MM-DD"

            request_main = request.GET.get("main", False)
            perms, main = get_character(request, character_id)
            entitys = get_main_and_alts_ids_corporations(request)

            if not perms:
                context = {
                    "error_title": "Permission Denied",
                    "error_message": "You don't have permission to view this character's ledger.",
                }
                return render(
                    request, "ledger/modals/information/error.html", context, status=403
                )

            alts = get_alts_queryset(main)
            chars_list = [char.character_id for char in alts]

            overall_mode = character_id == 0 or request_main
            current_date = timezone.now()

            try:
                if overall_mode:
                    linked_char = EveCharacter.objects.filter(
                        character_id__in=chars_list,
                    )
                else:
                    linked_char = [
                        EveCharacter.objects.get(
                            character_id=character_id,
                        )
                    ]

            except EveCharacter.DoesNotExist:
                context = {
                    "error_title": "403 Error",
                    "error_message": "Character not found.",
                }
                return render(
                    request, "ledger/modals/information/error.html", context, status=403
                )

            # Create the Ledger
            ledger_data = TemplateData(
                request=request,
                main=main,
                date=date_obj,
                view=view,
                corporations_ids=entitys,
                current_date=current_date,
            )
            ledger = TemplateProcess(linked_char, ledger_data, overall_mode)
            context = {
                "character": ledger.character_template(),
                "mode": "Ratting",
            }

            return render(
                request,
                "ledger/modals/information/view_character_content.html",
                context,
            )

        @api.get(
            "{entity_type}/{entity_id}/{main_id}/template/date/{date}/view/{view}/",
            response={200: list[schema.CharacterLedgerTemplate], 403: str},
            tags=self.tags,
        )
        # pylint: disable=too-many-positional-arguments, too-many-locals
        def get_ledger_corp_ally_details_information(
            request,
            entity_type: str,
            entity_id: int,
            main_id: int,
            date: str,
            view: str,
            corp: bool = False,
        ):
            try:
                date_obj = datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                return 403, "Invalid Date format. Use YYYY-MM-DD"

            perm, entitys = ledger_api_process(request, entity_type, entity_id)
            corp_view = False

            if not perm:
                context = {
                    "error_title": "Permission Denied",
                    "error_message": "You don't have permission to view this corporation's ledger.",
                }
                return render(
                    request, "ledger/modals/information/error.html", context, status=403
                )

            current_date = timezone.now()

            if entity_id == main_id:
                corp_view = True

            _, char = get_character(request, main_id, corp=corp_view)

            if char is None:
                context = {
                    "error_title": "403 Error",
                    "error_message": "Entity not found.",
                }
                return render(
                    request, "ledger/modals/information/error.html", context, status=403
                )

            overall_mode = main_id == 0

            if overall_mode:
                chars_list = get_journal_entitys(
                    date=date_obj, view=view, corporations=entitys
                )
                linked_char = EveEntity.objects.filter(
                    eve_id__in=chars_list,
                )
            elif corp:
                linked_char = EveEntity.objects.filter(
                    eve_id__in=[main_id],
                )
                overall_mode = True
            else:
                linked_char = get_corp_alts_queryset(char, corporations=None)

            # Create the Ledger
            ledger_data = TemplateData(
                request=request,
                main=char,
                date=date_obj,
                view=view,
                corporations_ids=entitys,
                current_date=current_date,
            )
            ledger = TemplateProcess(linked_char, ledger_data, overall_mode)
            context = {
                "character": ledger.corporation_template(),
                "mode": "TAX",
            }

            return render(
                request,
                "ledger/modals/information/view_character_content.html",
                context,
            )



from http import HTTPStatus
from unittest.mock import MagicMock, Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import ObjectDoesNotExist
from django.test import RequestFactory, TestCase
from django.urls import reverse

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.models.planetary import CharacterPlanetDetails
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.views.character.planetary import planetary_ledger, switch_alarm

MODULE_PATH = "ledger.views.character.planetary"


class CharPlaneterayTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user2, cls.character_ownership = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )

    def test_planetary_ledger_view(self):
        request = self.factory.get(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )
        request.user = self.user
        response = planetary_ledger(request, 0)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    @patch(MODULE_PATH + ".CharacterPlanetDetails")
    def test_switch_alarm_valid(
        self, mock_character_planet_details, mock_get_character, mock_messages
    ):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 1001})
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (True, MagicMock(character_id=1001))
        mock_character_planet_details.objects.filter.return_value = [
            MagicMock(notification=False)
        ]

        response = switch_alarm(request, [1001], 1)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    @patch(MODULE_PATH + ".CharacterPlanetDetails")
    def test_switch_alarm_all_planets(
        self, mock_character_planet_details, mock_get_character, mock_messages
    ):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 1001})
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (True, MagicMock(character_id=1001))
        mock_character_planet_details.objects.filter.return_value = [
            MagicMock(notification=False)
        ]

        response = switch_alarm(request, [1001], 0)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    @patch(MODULE_PATH + ".CharacterPlanetDetails")
    def test_switch_alarm_all_planets_and_alts(
        self, mock_character_planet_details, mock_get_character, mock_messages
    ):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 1001})
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (True, MagicMock(character_id=1001))
        mock_character_planet_details.objects.filter.return_value = [
            MagicMock(notification=False)
        ]

        response = switch_alarm(request, 0, 0)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    @patch(MODULE_PATH + ".CharacterPlanetDetails")
    def test_switch_alarm_invalid_character(
        self, mock_character_planet_details, mock_get_character, mock_messages
    ):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (False, None)

        response = switch_alarm(request, [9999], 1)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    @patch(MODULE_PATH + ".CharacterPlanetDetails")
    def test_switch_alarm_invalid_planet(
        self, mock_character_planet_details, mock_get_character, mock_messages
    ):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 1001})
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (True, MagicMock(character_id=1001))
        mock_character_planet_details.side_effect = CharacterPlanetDetails.DoesNotExist

        response = switch_alarm(request, [1001], 9999)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    def test_switch_alarm_no_planets(self, mock_get_character, mock_messages):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 1001})
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (True, MagicMock(character_id=1001))

        response = switch_alarm(request, [1001], 9999)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".get_character")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.filter")
    def test_switch_alarm_no_permission(
        self, mock_character_planet_details, mock_get_character, mock_messages
    ):
        request = self.factory.post(
            reverse("ledger:planetary_ledger", kwargs={"character_pk": 1001})
        )
        request.user = self.user2
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        mock_get_character.return_value = (False, None)

        response = switch_alarm(request, [1001], 1)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:planetary_ledger", kwargs={"character_pk": 0})
        )
